import anyio
from pathlib import Path
from runpy import run_path
from panther.configs import config
from panther.request import Request
from panther.response import Response
from panther.exceptions import APIException
from panther.utils import read_body, send_404, send_405, send_204
""" We can't import logger on the top cause it needs config['base_dir'] ans its fill in __init__ """


class Panther:

    def __init__(self, name):
        import os
        os.system('clear')
        self.base_dir = Path(name).resolve().parent
        config['base_dir'] = self.base_dir
        self.panther_dir = Path(__file__).parent
        self.load_configs()
        del os

    async def run(self, scope, receive, send):
        from panther.logger import logger, monitoring
        # Read Body & Create Request
        body = await read_body(receive)
        request = Request(scope=scope, body=body)
        # TODO: put user_ip, method to Request

        # TODO: put access_log to config

        # Access Log
        # TODO: use this log as a middleware so can have a response status too.(we should refactor the structure for this)
        monitoring.info(f"[{scope['method']}] {scope['path']} | {scope['client'][0]}:{scope['client'][1]}")

        # TODO: pass request.path to find_endpoint instead of scope[]
        # Find Endpoint
        endpoint = self.find_endpoint(path=scope['path'])
        if endpoint is None:
            return await send_404(send)

        # Check Endpoint Method
        if endpoint.__module__ != 'panther.app':
            raise TypeError(f'You have to use API decorator on {endpoint.__module__}.{endpoint.__name__}()')
        endpoint_method = endpoint.__qualname__.split('.')[1].upper()
        if endpoint_method != scope['method']:
            return await send_405(send)

        # Call 'Before' Middlewares
        for middleware in config['middlewares']:
            request = await middleware.before(request=request)

        # Call Endpoint
        try:
            response = await endpoint(request=request)
        except APIException as e:
            response = Response(
                data=e.detail if isinstance(e.detail, dict) else {'detail': e.detail},
                status_code=e.status_code
            )

        if not isinstance(response, Response):
            return logger.error(f"Response Should Be Instance Of 'Response'.")

        # Call 'After' Middleware
        config['middlewares'].reverse()
        for middleware in config['middlewares']:
            response = await middleware.after(response=response)

        # Return Response
        if response._data is None or response.status_code == 204:
            return await send_204(send)

        await send({
            'type': 'http.response.start',
            'status': response.status_code,
            'headers': [
                [b'content-type', b'application/json'],
            ],
        })
        await send({'type': 'http.response.body', 'body': response.data})

    async def __call__(self, scope, receive, send) -> None:
        # await self.run(scope, receive, send)
        async with anyio.create_task_group() as task_group:
            task_group.start_soon(self.run, scope, receive, send)
            # await anyio.to_thread.run_sync(self.run, scope, receive, send)
        # if self.exc_info is not None:
        #     raise self.exc_info[0].with_traceback(self.exc_info[1], self.exc_info[2])

        # with ProcessPoolExecutor() as e:
        #     e.submit(self.run, scope, receive, send)

    def load_configs(self) -> None:
        from panther.logger import logger
        logger.debug(f'Base Directory: {self.base_dir}')

        # Check Configs
        self.check_configs()
        config['debug'] = self.settings.get('DEBUG', config['debug'])

        # Collect Middlewares
        self.collect_middlewares()
        # Check & Collect URLs
        # check_urls should be the last call in load_configs, because it will read all files and load them.
        urls = self.check_urls()
        self.collect_urls('', urls)
        logger.debug('Configs Loaded.')

    def check_configs(self):
        from panther.logger import logger

        try:
            configs_path = self.base_dir / 'core/configs.py'
            self.settings = run_path(str(configs_path))
        except FileNotFoundError:
            return logger.critical('core/configs.py Not Found.')

        # URLs
        if 'URLs' not in self.settings:
            return logger.critical("configs.py Does Not Have 'URLs'")

    def check_urls(self) -> dict:
        from panther.logger import logger

        urls_path = self.settings['URLs']
        try:
            full_urls_path = self.base_dir / urls_path
            urls_dict = run_path(str(full_urls_path))['urls']
        except FileNotFoundError:
            return logger.critical("Couldn't Open 'URLs' Address.")
        except KeyError:
            return logger.critical("'URLs' Address Does Not Have 'urls'")
        if not isinstance(urls_dict, dict):
            return logger.critical("'urls' Of URLs Is Not dict.")
        return urls_dict

    def collect_urls(self, pre_url, urls):
        from panther.logger import logger

        for url, endpoint in urls.items():
            if endpoint is ...:
                logger.error(f"URL Can't Point To Ellipsis. ('{pre_url}{url}' -> ...)")
            if endpoint is None:
                logger.error(f"URL Can't Point To None. ('{pre_url}{url}' -> None)")

            if isinstance(endpoint, dict):
                self.collect_urls(f'{pre_url}/{url}', endpoint)
            else:
                config['urls'][f'{pre_url}{url}'] = endpoint
        return urls

    def collect_middlewares(self):
        from panther.logger import logger

        # TODO: use importlib
        # TODO: is sub instance of BaseMiddleware
        _middlewares = self.settings['Middlewares']
        for _middleware in _middlewares:
            if _middleware[0].split('/')[0] == 'panther':
                _middleware_name = _middleware[0].split('/')[-1]
                if _middleware_name == 'db.py':
                    config['db_engine'] = _middleware[1]['url'].split(':')[0]
                    from panther.middlewares.db import Middleware
                elif _middleware_name == 'redis.py':
                    from panther.middlewares.redis import Middleware
                else:
                    logger.error(f'{_middleware[0]} Does Not Found.')
                    continue
            else:
                # TODO: Import From Example (Custom Middleware)
                ...

            config['middlewares'].append(Middleware(**_middleware[1]))

    def find_endpoint(self, path):
        # TODO: Fix it later, it does not support root url or something like ''
        for url in config['urls']:
            if path == url:
                return config['urls'][url]



from typing import TypedDict
from pathlib import Path


class Config(TypedDict):
    base_dir: Path
    debug: bool
    urls: dict
    middlewares: list
    db_engine: str


config: Config = {
    'base_dir': Path(),
    'debug': False,
    'urls': {},
    'middlewares': [],
    'db_engine': '',
}



import orjson as json
from types import NoneType


class Response:
    def __init__(self, data: dict | list | tuple | str | bool = None, status_code: int = 200):
        """
        :param data: should be dict, list, tuple, str, bool
        :param status_code: should be int
        """
        if type(data) not in [dict, list, tuple, str, bool, NoneType]:
            raise TypeError(f"Response data can't be '{type(data).__name__}'")
        self._status_code = status_code
        self._data = data
        # TODO: Add Header To Response

    @property
    def status_code(self) -> int:
        if isinstance(self._status_code, int):
            return self._status_code
        else:
            raise TypeError(f"Response 'status_code' Should Be int. ('{self._status_code}' -> {type(self._status_code)})")

    @property
    def data(self) -> bytes:
        return json.dumps(self._data)
        # if isinstance(self._data, dict) or isinstance(self._data, list) or isinstance(self._data, tuple):
        #     return json.dumps(self._data)
        # else:  # str, bool
        #     return json.dumps({'detail': self._data})

    def set_data(self, data: dict) -> None:
        self._data = data




import os
import sys
from pathlib import Path
from subprocess import Popen
from watchfiles import watch
from rich import print as rprint
from panther.cli.template import Template

logo = r"""│    ____                 __    __                         │
│   /\  _`\              /\ \__/\ \                        │
│   \ \ \L\ \ __      ___\ \ ,_\ \ \___      __   _ __     │
│    \ \ ,__/'__`\  /' _ `\ \ \/\ \  _ `\  /'__`\/\`'__\   │
│     \ \ \/\ \L\.\_/\ \/\ \ \ \_\ \ \ \ \/\  __/\ \ \/    │
│      \ \_\ \__/.\_\ \_\ \_\ \__\\ \_\ \_\ \____\\ \_\    │
│       \/_/\/__/\/_/\/_/\/_/\/__/ \/_/\/_/\/____/ \/_/    │
"""

help_message = f"""╭{58 * '─'}╮
{logo}│{58 * ' '}│
│                                                          │
│   usage:                                                 │
│       - panther create <project_name>                    │
│           Create your project in current directory       │
│                                                          │
│       - panther run [--reload]                           │
│           Run your project with uvicorn                  │
│                                                          │
│       - panther shell                                    │
│           Run interactive python shell                   │
│                                                          │
│       - panther monitor                                  │
│           Show the monitor :)                            │
│                                                          │
│       - panther [--help | -h | help]                     │
│           Show this message and exit                     │
╰{58 * '─'}╯
"""


def error(message: str) -> None:
    print(f'Error: {message}\n\nuse panther -h for more help')


def create(args: list):
    base_dir = Path(__name__).resolve().parent

    if len(args) == 0:
        return error('Not Enough Parameters.')
    project_name = args[0]

    # TODO: Add Custom BaseDirectory

    if os.path.isdir(project_name):
        return error(f'"{project_name}" Directory Already Exists.')
    os.makedirs(project_name)

    for file_name, data in Template.items():
        if isinstance(data, dict):
            sub_directory = f'{project_name}/{file_name}'
            os.makedirs(sub_directory)
            for sub_file_name, sub_data in data.items():
                file_path = f'{sub_directory}/{sub_file_name}'
                with open(file_path, 'x') as file:
                    file.write(sub_data)
        else:
            if file_name == '.env':
                data = data.replace('{DATABASE_NAME}', project_name.lower())

            file_path = f'{project_name}/{file_name}'
            with open(file_path, 'x') as file:
                file.write(data)


def run(args) -> None:
    command = ['uvicorn', 'main:app', '--no-access-log']
    command.extend(args)
    sp = Popen(command)
    try:
        sp.wait()
    except KeyboardInterrupt:
        sp.kill()


def shell() -> None:
    os.system('bpython')


def monitor() -> None:
    # TODO: Is it only watch logs/monitoring.log or the whole directory ?
    try:
        from rich import box
        from rich.table import Table
        from rich.live import Live
        from rich.layout import Layout
        from datetime import datetime

        from rich import box
        from rich.align import Align
        from rich.console import Console, Group
        from rich.layout import Layout
        from rich.panel import Panel
        from rich.progress import Progress, SpinnerColumn, BarColumn, TextColumn
        from rich.syntax import Syntax
        from rich.table import Table
        from rich.text import Text

        table = Table(box=box.MINIMAL_DOUBLE_HEAD)
        table.add_column('Datetime', justify='right', style='magenta', no_wrap=True)
        table.add_column('Message', style='cyan')
        table.add_column('Response Time', justify='right', style='blue')

        with open('logs/monitoring.log', 'r') as f:
            f.readlines()

            layout = Layout(name='root')
            layout['root'].update(
                Panel(
                    Align.center(Group(table)),
                    box=box.ROUNDED,
                    padding=(1, 2),
                    title='Monitoring',
                    border_style='bright_blue',
                )
            )

            with Live(layout, auto_refresh=False, vertical_overflow='visible', screen=True) as live:
                for _ in watch('logs/monitoring.log'):
                    data = f.readline().split('|')
                    table.add_row(*data)
                    live.update(layout)
                    live.refresh()

    except FileNotFoundError:
        error("Monitor Log File Does Not Exists.\n\nHint: Make sure 'Monitor' is True in 'core/configs' "
              "or you are in a correct directory.")
    except KeyboardInterrupt:
        pass


def start() -> None:
    if len(sys.argv) == 1 or sys.argv[1] in ['help', '-h', '--help']:
        rprint(help_message)
        return

    if sys.argv[1] == 'create':
        create(sys.argv[2:])
        print('Project Created Successfully.')
    elif sys.argv[1] == 'run':
        run(sys.argv[2:])
    elif sys.argv[1] == 'shell':
        shell()
    elif sys.argv[1] == 'monitor':
        monitor()

    else:
        error('Invalid Arguments.')



from panther.configs import config
from panther.logger import logger
from bson import ObjectId, errors
from time import perf_counter
from typing import Union


def query_logger(func):
    def log(*args, **kwargs):
        if config['debug'] is False:
            return func(*args, **kwargs)
        start = perf_counter()
        response = func(*args, **kwargs)
        end = perf_counter()
        class_name = args[0].__name__ if hasattr(args[0], '__name__') else args[0].__class__.__name__
        logger.info(f'\033[1mQuery -->\033[0m  {class_name}.{func.__name__}() --> {(end - start) * 1_000:.2} ms')
        # logger.info(f'Query --> {class_name}.{func.__name__}() --> {(end - start) * 1_000:.2} ms')
        return response
    return log


def to_object_id(_id: Union[ObjectId, str]) -> ObjectId:
    if isinstance(_id, ObjectId):
        return _id
    try:
        return ObjectId(_id)
    except errors.InvalidId:
        raise



from orjson.orjson import JSONDecodeError

from pydantic import ValidationError
from pydantic.main import BaseModel

from panther.request import Request
from panther.response import Response
from panther.exceptions import APIException


class API:

    @classmethod
    def validate_input(cls, request: Request, input_model):
        if input_model:
            try:
                validated_data = input_model(**request.data)
                request.set_data(validated_data)
            except ValidationError as validation_error:
                error = {e['loc'][0]: e['msg'] for e in validation_error.errors()}
                raise APIException(status_code=400, detail=error)
            except JSONDecodeError:
                raise APIException(status_code=400, detail={'detail': 'JSON Decode Error'})

    @classmethod
    def clean_output(cls, data, output_model):
        if data is None or output_model is None:
            return data
        if isinstance(data, dict):
            _data = output_model(**data).dict()
        elif issubclass(type(data), BaseModel):
            _data = output_model(**data.dict()).dict()
        elif isinstance(data, list) or isinstance(data, tuple):
            if len(data) == 0:
                return data
            if isinstance(data[0], BaseModel):
                _data = [output_model(**d.dict()).dict() for d in data]
            else:
                _data = [output_model(**d).dict() for d in data]
        elif isinstance(data, str) or isinstance(data, bool) or isinstance(data, set):
            raise TypeError('Type of Response data is not match with output_model. '
                            '\n*hint: You may want to pass None to output_model')
        else:
            raise TypeError(f"Type of Response 'data' is not valid.")
        return _data

    @classmethod
    def post(cls, input_model=None, output_model=None):
        def decorator(func):
            async def wrapper(*args, **kwargs):
                request: Request = kwargs['request']
                cls.validate_input(request=request, input_model=input_model)
                response = await func(request) if Request in func.__annotations__.values() else await func()
                if not isinstance(response, Response):
                    response = Response(data=response)
                data = cls.clean_output(data=response._data, output_model=output_model)
                response.set_data(data)
                return response
            return wrapper
        return decorator

    @classmethod
    def put(cls, input_model=None, output_model=None):
        def decorator(func):
            async def wrapper(*args, **kwargs):
                request: Request = kwargs['request']
                cls.validate_input(request=request, input_model=input_model)
                response = await func(request) if Request in func.__annotations__.values() else await func()
                if not isinstance(response, Response):
                    response = Response(data=response)
                data = cls.clean_output(data=response._data, output_model=output_model)
                response.set_data(data)
                return response
            return wrapper
        return decorator

    @classmethod
    def get(cls, output_model=None):
        def decorator(func):
            async def wrapper(*args, **kwargs):
                request: Request = kwargs['request']
                response = await func(request) if Request in func.__annotations__.values() else await func()
                if not isinstance(response, Response):
                    response = Response(data=response)
                data = cls.clean_output(data=response._data, output_model=output_model)
                response.set_data(data)
                return response
            return wrapper
        return decorator

    @classmethod
    def delete(cls, output_model=None):
        def decorator(func):
            async def wrapper(*args, **kwargs):
                request: Request = kwargs['request']
                response = await func(request) if Request in func.__annotations__.values() else await func()
                if not isinstance(response, Response):
                    response = Response(data=response)
                data = cls.clean_output(data=response._data, output_model=output_model)
                response.set_data(data)
                return response
            return wrapper
        return decorator



from jose import JWTError, jwt
from datetime import datetime
from panther.configs import JWTConfig
from example.core.configs import JWTConfig
from panther.exceptions import CredentialsException


class JWTAuthentication:
    def __init__(self, token: str):
        self.token = token
        self.user_id = self.decode_jwt(token)

    def __str__(self):
        return f'Authentication(user_id={self.user_id}, token={self.token[:20]}...)'

    @staticmethod
    def encode_jwt(user_id: int) -> str:
        """ Encode JWT from user_id """
        expire = datetime.utcnow() + JWTConfig['TokenLifeTime']
        access_payload = {
            'token_type': 'access',
            'user_id': user_id,
            'exp': expire
        }
        return jwt.encode(access_payload, JWTConfig['Key'], algorithm=JWTConfig['Algorithm'])

    @staticmethod
    def decode_jwt(token: str) -> int:
        """ Decode JWT token to user_id (it can return multiple variable ... ) """
        try:
            payload = jwt.decode(token, JWTConfig['Key'], algorithms=[JWTConfig['Algorithm']])
            user_id: int = payload.get('user_id')
            if user_id is None:
                raise CredentialsException
        except JWTError:
            raise CredentialsException
        return user_id



from redis import Redis
from typing import Union
from tinydb import TinyDB
from pymongo import MongoClient
from pymongo.database import Database


class Singleton(object):
    _instances = {}

    def __new__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__new__(cls, *args, **kwargs)
        return cls._instances[cls]


class DBSession(Singleton):
    _session: Union[TinyDB, Database]
    _client: MongoClient
    _name: str

    def __init__(self, db_url: str | None = None):
        if db_url:
            self._name = db_url[:db_url.find(':')]
            match self._name:
                case 'mongodb':
                    self._create_mongodb_session(db_url)
                case 'tinydb':
                    self._create_tinydb_session(db_url[9:])
                case _:
                    # TODO: self._name does not have a last character if only path passed
                    raise ValueError(f'We are support {self._name} Database yet')

    @property
    def session(self) -> Union[TinyDB, Database]:
        return self._session

    @property
    def name(self) -> str:
        return self._name

    def _create_mongodb_session(self, db_url: str) -> None:
        self._client = MongoClient(db_url)
        self._session: Database = self._client.get_database()

    def _create_tinydb_session(self, db_url: str):
        self._session: TinyDB = TinyDB(db_url)

    def close(self):
        if self._name == 'mongodb':
            self._client.close()
        else:
            self._session.close()


class RedisConnection(Singleton, Redis):
    def __init__(self, host: str | None = None, port: int | None = None, **kwargs):
        if host and port:
            super().__init__(host=host, port=port, **kwargs)


db: DBSession = DBSession()
redis: Redis = RedisConnection()





from panther.logger import logger
from panther.db.connection import db
from panther.db.utils import query_logger


class BaseTinyDBQuery:

    @classmethod
    @query_logger
    def get_one(cls, _data: dict = None, /, **kwargs):
        from tinydb import Query
        if _data is None:
            _data = {}
        _data = {k: v for k, v in _data.items() if v not in [None]}
        kwargs = {k: v for k, v in kwargs.items() if v not in [None]}
        result = db.session.search(Query().fragment(_data | kwargs))
        if len(result) < 1:
            return None
        else:
            return result[0]

    @classmethod
    @query_logger
    def count(cls, _data: dict = None, /, **kwargs) -> None:
        logger.critical('count() is not supported while using TinyDB.')

    @classmethod
    @query_logger
    def list(cls, _data: dict = None, /, **kwargs):
        from tinydb import Query
        if _data is None:
            _data = {}
        _data = {k: v for k, v in _data.items() if v not in [None]}
        kwargs = {k: v for k, v in kwargs.items() if v not in [None]}
        return db.session.search(Query().fragment(_data | kwargs))

    @classmethod
    @query_logger
    def create(cls, _data: dict = None, **kwargs) -> int:
        return db.session.insert(_data | kwargs)

    @query_logger
    def delete(self) -> None:
        logger.critical('delete() is not supported while using TinyDB.')

    @classmethod
    @query_logger
    def delete_one(cls, **kwargs) -> None:
        logger.critical('delete_one() is not supported while using TinyDB.')

    @classmethod
    @query_logger
    def delete_many(cls, **kwargs) -> None:
        logger.critical('delete_many() is not supported while using TinyDB.')

    @query_logger
    def update(self, **kwargs) -> None:
        logger.critical('update() is not supported while using TinyDB.')

    @classmethod
    @query_logger
    def update_one(cls, _filter, _data: dict = None, /, **kwargs) -> None:
        logger.critical('update_one() is not supported while using TinyDB.')

    @classmethod
    @query_logger
    def update_many(cls, _filter, **kwargs) -> None:
        logger.critical('update_many() is not supported while using TinyDB.')

    @classmethod
    @query_logger
    def increment(cls, _filter, **kwargs) -> None:
        logger.critical('increment() is not supported while using TinyDB.')



import orjson as json


async def send_404(send):
    # TODO: Work On This Func
    await send({
        'type': 'http.response.start',
        'status': 404,
        'headers': [
            [b'content-type', b'application/json'],
        ],
    })
    return await send({
        'type': 'http.response.body',
        'body': json.dumps({'detail': 'Not Found.'}),
    })


async def send_204(send):
    await send({
        'type': 'http.response.start',
        'status': 204,
        'headers': [
            [b'content-type', b'application/json'],
        ],
    })
    return await send({
        'type': 'http.response.body',
    })


async def send_405(send):
    # TODO: Work On This Func
    await send({
        'type': 'http.response.start',
        'status': 405,
        'headers': [
            [b'content-type', b'application/json'],
        ],
    })
    return await send({
        'type': 'http.response.body',
        'body': json.dumps({'detail': 'Method Not Allowed.'}),
    })


async def read_body(receive) -> bytes:
    """
    Read and return the entire body from an incoming ASGI message.
    """
    body = b''
    more_body = True

    while more_body:
        message = await receive()
        body += message.get('body', b'')
        more_body = message.get('more_body', False)

    return body




from panther.logger import logger
from panther.middlewares.base import BaseMiddleware
from panther.db.connection import RedisConnection
from panther.response import Response
from panther.request import Request


class Middleware(BaseMiddleware):

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.validate_host()
        self.validate_port()

    def validate_host(self):
        _host = self.kwargs.get('host')
        if _host:
            if False:  # TODO: should be valid ip or domain
                logger.critical('Redis Host Is Not Valid.')
        else:
            self.kwargs['host'] = '127.0.0.1'

    def validate_port(self):
        _port = self.kwargs.get('port')
        if _port:
            if not isinstance(_port, int):
                logger.critical('Redis Port Is Not Valid.')
        else:
            self.kwargs['port'] = '6379'

    async def before(self, request: Request):
        self.redis = RedisConnection(**self.kwargs)
        return request

    async def after(self, response: Response):
        self.redis.close()
        return response





from bson.objectid import ObjectId
from panther.configs import config
from panther.db.queries.mongo_queries import BaseMongoDBQuery
from panther.db.queries.tinydb_queries import BaseTinyDBQuery


BaseQuery = BaseTinyDBQuery if config['db_engine'] == 'tinydb' else BaseMongoDBQuery

__all__ = (
    'Query',
)


class Query(BaseQuery):

    @classmethod
    def get_one(cls, _data: dict = None, /, **kwargs):
        """
        example:
            if MongoDB:
                >>> User.get_one(id=id)
            elif TinyDB:
                >>> User.get_one(id=id)
        """
        return super().get_one(_data, **kwargs)

    @classmethod
    def count(cls, _data: dict = None, /, **kwargs) -> int:
        """
        example:
            if MongoDB:
                >>> User.count(name='ali', age=24)
            elif TinyDB:
                ...
        """
        return super().count(_data, **kwargs)

    @classmethod
    def list(cls, _data: dict = None, /, **kwargs):
        """
        example:
            if MongoDB:
                >>> User.list({'state': {'$eq': 'Confirmed'}}, name='ali', age=24)
            elif TinyDB:
                >>> User.list(name='ali', age=24)
        """
        return super().list(_data, **kwargs)

    @classmethod
    def create(cls, _data: dict = None, **kwargs) -> ObjectId:
        """
        example:
            if MongoDB:
                >>> User.create(name='ali', age=24, ...)
            elif TinyDB:
                >>> User.create(name='ali', age=24, ...)
        """
        return super().create(_data, **kwargs)

    def delete(self) -> bool:
        """
        example:
            if MongoDB:
                >>> user = User.get_one(name='ali')
                >>> user.delete()
            elif TinyDB:
                ...
        """
        return super().delete()

    @classmethod
    def delete_one(cls, **kwargs) -> bool:
        """
        example:
            if MongoDB:
                >>> User.delete_one(id=_id)
            elif TinyDB:
                ...
        """
        return super().delete_one(**kwargs)

    @classmethod
    def delete_many(cls, **kwargs) -> int:
        """
        example:
            if MongoDB:
                >>> User.delete_many(name='ali')
            elif TinyDB:
                ...
        """
        return super().delete_many(**kwargs)

    def update(self, **kwargs) -> dict:
        """
        example:
            if MongoDB:
                >>> user = User.get_one(name='ali')
                >>> user.update(name='tom')
            elif TinyDB:
                ...
        """
        return super().update(**kwargs)

    @classmethod
    def update_one(cls, _filter, _data: dict = None, /, **kwargs) -> dict:
        """
        example:
            if MongoDB:
                >>> User.update_one({'id': _id}, name='ali')
            elif TinyDB:
                ...
        """
        return super().update_one(_filter, _data, **kwargs)

    @classmethod
    def update_many(cls, _filter, **kwargs) -> dict:
        """
        example:
            if MongoDB:
                >>> User.update_many({'name': 'mohsen'}, name='ali')
            elif TinyDB:
                ...
        """
        return super().update_many(_filter, **kwargs)

    @classmethod
    def increment(cls, _filter, **kwargs):
        """
        example:
            if MongoDB:
                >>> User.increment({'priority': {'$gt': ad.priority}}, score=1)
            elif TinyDB:
                ...
        * it will increment score by 1
        """
        return super().increment(_filter, **kwargs)



from dataclasses import dataclass


@dataclass
class APIException(Exception):
    detail: str | dict | list = ''
    status_code: int = 400


class UserNotFound(APIException):
    detail = 'not found'
    status_code = 403


class MissingSessionError(APIException):
    """
    Exception raised for when the user tries to access a database session before it is created.
    """
    detail = "MissingSessionError"
    status_code = 402


class SessionNotInitialisedError(APIException):
    """
    Exception raised when the user creates a new DB session without first initialising it.
    """
    detail = "Session not initialised!"
    status_code = 401



import orjson as json
from dataclasses import dataclass


@dataclass(frozen=True)
class Headers:
    accept_encoding: str
    content_length: int
    authorization: str
    content_type: str
    user_agent: str
    connection: str
    accept: str
    host: str


class Request:
    def __init__(self, scope: dict, body: bytes):
        """
        {'type': 'http', 'asgi': {'version': '3.0', 'spec_version': '2.3'},
        'http_version': '1.1', 'server': ('127.0.0.1', 8000), 'client': ('127.0.0.1', 35064),
        'scheme': 'http', 'root_path': '', 'headers': [
            (b'content-type', b'application/json'),
            (b'user-agent', b'PostmanRuntime/7.29.2'),
            (b'accept', b'*/*'),
            (b'postman-token', b'3e78fbf3-df2f-41bd-bedc-cf6027fa4fe6'),
            (b'host', b'127.0.0.1:8000'),
            (b'accept-encoding', b'gzip, deflate, br'),
            (b'connection', b'keep-alive'),
            (b'content-length', b'55')
        ],
        'method': 'GET', 'path': '/list/', 'raw_path': b'/list/', 'query_string': b''}
        """
        self.scope = scope
        self._body = body
        self._data = None

    @property
    def headers(self):
        _headers = {header[0].decode('utf-8'): header[1].decode('utf-8') for header in self.scope['headers']}

        # logger.debug(_headers)
        return Headers(
            accept_encoding=_headers.pop('accept-encoding', None),
            content_length=_headers.pop('content_length', None),
            authorization=_headers.pop('authorization', None),
            content_type=_headers.pop('content-type', None),
            user_agent=_headers.pop('user-agent', None),
            connection=_headers.pop('connection', None),
            accept=_headers.pop('accept', None),
            host=_headers.pop('host', None),
            # TODO: Others ...
        )

    @property
    def query_params(self):
        query_string = self.scope['query_string'].decode('utf-8').split('&')
        params = {}
        for param in query_string:
            k, *_, v = param.split('=')
            params[k] = v
        return params

    @property
    def method(self):
        return self.scope['method']

    @property
    def path(self):
        return self.scope['path']

    @property
    def server(self):
        return self.scope['server']

    @property
    def client(self):
        return self.scope['client']

    @property
    def http_version(self):
        return self.scope['http_version']

    @property
    def scheme(self):
        return self.scope['scheme']

    @property
    def data(self):
        from panther.logger import logger

        if self._data:
            return self._data

        body = self._body.decode('utf-8') or {}
        if self.headers.content_type is None:
            # logger.error(f'request content-type is None.')
            _data = body
        elif self.headers.content_type == 'application/json':
            _data = json.loads(body)
        elif self.headers.content_type[:19] == 'multipart/form-data':
            # TODO: Handle Multipart Form Data
            logger.error(f"We Don't Handle Multipart Request Yet.")
            _data = None
        else:
            logger.error(f'{self.headers.content_type} Is Not Supported.')
            _data = None

        # return {'id': 1, 'username': 'ali', 'password': '1123'}  # TODO: For Testing ...
        return _data

    def set_data(self, data) -> None:
        self._data = data



from datetime import datetime


apis_py = """from panther.app import API
from panther.request import Request
from panther.response import Response

from app.models import User
from app.serializers import UserInputSerializer, UserOutputSerializer


@API.post(input_model=UserInputSerializer)
async def create_user(request: Request):
    obj: UserInputSerializer = request.data
    User.create(username=obj.username, password=obj.password)
    return Response(status_code=201)


@API.get(output_model=UserOutputSerializer)
async def get_users():
    users = User.list()
    return Response(data=users)

"""

models_py = """from panther.db import BaseModel

class User(BaseModel):
    username: str
    password: str

"""

serializers_py = """from pydantic import BaseModel, constr


class UserInputSerializer(BaseModel):
    username: str
    password: constr(min_length=8)


class UserOutputSerializer(BaseModel):
    id: str
    username: str

"""

app_urls_py = """from app.apis import create_user, get_users

urls = {
    'create/': create_user,
    'list/': get_users,
}

"""

configs_py = """\"""
Generated by Panther on %s
\"""

from pathlib import Path
from dotenv import dotenv_values


DEBUG = True 
BASE_DIR = Path(__name__).resolve().parent
env = dotenv_values(BASE_DIR / '.env')

DB_NAME = env['DB_NAME']
SECRET_KEY = env['SECRET_KEY']

Middlewares = [
    ('panther/middlewares/db.py', {'url': f'tinydb://{BASE_DIR}/{DB_NAME}.json'}),
]

URLs = 'core/urls.py'
""" % datetime.now().date().isoformat()

env = """
SECRET_KEY = 'THIS_IS_THE_SECRET_SECRET_KEY'

DB_NAME = '{DATABASE_NAME}'

"""

main_py = """from panther import Panther

app = Panther(__name__)

"""

urls_py = """from app.urls import urls as app_urls

urls = {
    '': app_urls, 
}

"""

git_ignore = """__pycache__/
.idea/
.venv/
__pycache__/
.env

"""

Template = {
    'app': {
        'apis.py': apis_py,
        'models.py': models_py,
        'serializers.py': serializers_py,
        'urls.py': app_urls_py,
    },
    'core': {
        'configs.py': configs_py,
        'urls.py': urls_py,
    },
    'main.py': main_py,
    '.env': env,
    '.gitignore': git_ignore,
}

# TODO: Add .gitignore

# TODO: Add core/middlewares.py to Template



import os
import re
from setuptools import setup


def panther_version() -> str:
    with open(os.path.join('panther/__init__.py')) as f:
        return re.search("__version__ = ['\"]([^'\"]+)['\"]", f.read()).group(1)


_version = panther_version()
_long_description = open('README.md').read()

setup(
    name='panther',
    version=_version,
    python_requires='>=3.10',
    author='Ali RajabNezhad',
    author_email='alirn76@yahoo.com',
    keywords='fast friendly web framework',
    url='https://github.com/alirn76/panther',
    description='Fast, Friendly Python Web Framework',
    long_description=_long_description,
    long_description_content_type='text/markdown',
    include_package_data=True,
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.10',
    ],
    entry_points={
        'console_scripts': ['panther=panther.cli.main:start'],
    },
    package_data={
        'panther': ['cli/*']
    },
    install_requires=[
        'uvicorn[standard]',
        'orjson>=3.7.12',
        'pydantic>=1.9.2',
        'redis>=4.3.4',
        'pymongo>=4.2.0',
        'bpython>=0.23',
        'tinydb>=4.7.0',
        'rich>=12.5.1',
    ],
)



import bson
from typing import Tuple
from typing import Type, TypeVar
from panther.db.utils import query_logger, to_object_id
from panther.db.connection import db  # # # Do Not Delete This Import (Used in eval)


# TODO: Not sure about this bounding
T = TypeVar('T', bound='BaseMongoDBQuery')


class BaseMongoDBQuery:

    @classmethod
    @query_logger
    def get_one(cls: Type[T], _data: dict = None, /, **kwargs) -> T:
        if _data is None:
            _data = {}
        if '_id' in _data:
            _data['_id'] = to_object_id(_data['_id'])
        if '_id' in kwargs:
            kwargs['_id'] = to_object_id(kwargs['_id'])
        obj = eval(f'db.session.{cls.__name__}.find_one(_data | kwargs)')
        return cls(**obj) if obj else None

    @classmethod
    @query_logger
    def count(cls, _data: dict = None, /, **kwargs) -> int:
        if _data is None:
            _data = {}

        _data = {k: v for k, v in _data.items() if v not in [None]}
        kwargs = {k: v for k, v in kwargs.items() if v not in [None]}
        if '_id' in _data:
            _data['_id'] = to_object_id(_data['_id'])
        if '_id' in kwargs:
            kwargs['_id'] = to_object_id(kwargs['_id'])
        if 'id' in _data:
            _data['_id'] = to_object_id(_data['id'])
        if 'id' in kwargs:
            kwargs['_id'] = to_object_id(kwargs['id'])
        return eval(f'db.session.{cls.__name__}.count_documents(_data | kwargs)')

    @classmethod
    @query_logger
    def list(cls, _data: dict = None, /, **kwargs):
        if _data is None:
            _data = {}
        _data = {k: v for k, v in _data.items() if v not in [None]}
        kwargs = {k: v for k, v in kwargs.items() if v not in [None]}
        if '_id' in _data:
            _data['_id'] = to_object_id(_data['_id'])
        if '_id' in kwargs:
            kwargs['_id'] = to_object_id(kwargs['_id'])
        if 'id' in _data:
            _data['_id'] = to_object_id(_data['id'])
        if 'id' in kwargs:
            kwargs['_id'] = to_object_id(kwargs['id'])
        result = eval(f'db.session.{cls.__name__}.find(_data | kwargs)')
        return [cls(**obj) for obj in result]

    @classmethod
    @query_logger
    def create(cls, _data: dict = None, **kwargs) -> bson.objectid.ObjectId:
        return eval(f'db.session.{cls.__name__}.insert_one({_data or kwargs})').inserted_id

    @query_logger
    def delete(self) -> bool:
        _filter = {'_id': self._id}
        result = eval(f'db.session.{self.__class__.__name__}.delete_one(_filter)')
        return bool(result.deleted_count)

    @classmethod
    @query_logger
    def delete_one(cls, **kwargs) -> bool:
        if '_id' in kwargs:
            kwargs['_id'] = to_object_id(kwargs['_id'])
        if 'id' in kwargs:
            kwargs['_id'] = to_object_id(kwargs['id'])
        result = eval(f'db.session.{cls.__name__}.delete_one(_data)')
        return bool(result.deleted_count)

    @classmethod
    @query_logger
    def delete_many(cls, **kwargs) -> int:
        result = eval(f'db.session.{cls.__name__}.delete_many(kwargs)')
        return result.deleted_count

    @query_logger
    def update(self, _data: dict = None, **kwargs) -> dict:
        for field, value in (_data or kwargs).items():
            if hasattr(self, field):
                setattr(self, field, value)
        _filter = {'_id': self._id}
        _update = {'$set': kwargs}
        return eval(f'db.session.{self.__class__.__name__}.update_one(_filter, _update)')

    @classmethod
    @query_logger
    def update_one(cls, _filter, _data: dict = None, /, **kwargs) -> dict:
        if '_id' in _filter:
            _filter['_id'] = to_object_id(_filter['_id'])
        if 'id' in _filter:
            _filter['_id'] = to_object_id(_filter['id'])
        if _data is None:
            _data = {}
        _data = {k: v for k, v in _data.items() if v is not None}
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        _update = {'$set': kwargs}

        return eval(f'db.session.{cls.__name__}.update_one(_filter, _data | _update)')

    @classmethod
    @query_logger
    def update_many(cls, _filter, **kwargs) -> dict:
        _update = {'$set': kwargs}
        return eval(f'db.session.{cls.__name__}.update_many(_filter, _update)')

    @classmethod
    @query_logger
    def increment(cls, _filter, **kwargs):
        _update = {'$inc': kwargs}
        return eval(f'db.session.{cls.__name__}.update_many({_filter}, {_update})')

    @classmethod
    @query_logger
    def get_or_create(cls, **kwargs) -> Tuple[bool, any]:
        obj = cls.get_one(**kwargs)
        if obj:
            return False, obj
        else:
            return True, cls.create(**kwargs)



from bson import ObjectId
from pydantic import Field
from typing import Optional
from bson.errors import BSONError
from panther.db.queries import Query
from pydantic.main import BaseModel as PydanticBaseModel


class BsonObjectId(ObjectId):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, str):
            try:
                ObjectId(v)
            except BSONError:
                raise TypeError('Invalid ObjectId')
        elif not isinstance(v, ObjectId):
            raise TypeError('ObjectId required')
        return str(v)


class BaseModel(PydanticBaseModel, Query):
    id: Optional[BsonObjectId] = Field(alias='_id')

    @property
    def _id(self):
        return ObjectId(self.id) if self.id else None
